#define KERBEROS
/* Tastes like reauth, but no lumps */

#ifdef ULTRIX
#define LOGLEVEL LOG_INFO
#define LOGFACILITY LOG_INFO
#else
#define LOGLEVEL LOG_INFO
#define LOGFACILITY LOG_USER
#endif

#include <syslog.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <krb.h>
#include <afs/param.h>
#include <afs/auth.h>
#include <afs/cellconfig.h>
#include <afs/vice.h>
#include <afs/venus.h>

extern int errno;
int verbose;

void 
upcase (s)
     char *s;
{
  while (s[0])
    {
      s[0] = toupper (s[0]);
      s++;
    }
}

void 
downcase (s)
     char *s;
{
  while (s[0])
    {
      s[0] = tolower (s[0]);
      s++;
    }
}

int 
key_to_key (user, instance, realm, passwd, key)
     char *user, *instance, *realm, *passwd;
     C_Block key;
{
  memcpy (key, passwd, sizeof (des_cblock));
  return (0);
}

void 
usage ()
{
  fprintf (stderr, "usage: reauth [-v] [-nofork] [-srvtab <filename>] [-pw <filename>] <time> principal[.instance] [password]\n");
  exit (1);
}

static int 
logme (name, inst, password)
     char *name, *inst, *password;
{
  char *reason;
#ifdef KERBEROS
  char realm[REALM_SZ];
  int err;
  CREDENTIALS c;
  C_Block afskey, mitkey, use_as_key;
#ifndef NO_AFS
  struct ktc_principal aserver;
  struct ktc_principal aclient;
  struct ktc_token atoken, btoken;
#endif
#ifdef CYGNUS_V4
  key_proc_type authfunc = NULL;
#else
  key_proc_t authfunc = NULL;
#endif
#endif

  if (verbose)
    {
      if (*inst)
	{
	  syslog (LOGLEVEL, "Attempting reauthentication for %s.%s", name, inst);
	}
      else
	{
	  syslog (LOGLEVEL, "Attempting reauthentication for %s", name);
	}
    }				/* if verbose */

#ifdef NOISY
  printf ("Attempting Re-Authentication (%s", name);
  if (*inst)
    printf (".%s", inst);
  puts (")");
#endif

#ifdef KERBEROS
  (void) krb_get_lrealm (realm, 1);

  /* Generate the possible keys */
  ka_StringToKey (password, realm, afskey);
  des_string_to_key (password, mitkey);

#ifdef DEBUG
  printf ("AFS key: %x %x\n", *((unsigned long *) afskey) & 0xffffffff, *((unsigned long *) afskey + 1) & 0xffffffff);
  printf ("MIT key: %x %x\n", *((unsigned long *) mitkey) & 0xffffffff, *((unsigned long *) mitkey + 1) & 0xffffffff);
  printf ("key: %x %x\n", *((unsigned long *) use_as_key) & 0xffffffff, *((unsigned long *) use_as_key + 1) & 0xffffffff);
#endif

  if (krb_get_in_tkt (name, inst, realm, "krbtgt", realm, DEFAULT_TKT_LIFE, key_to_key, NULL, password) != 0)
    {
      if (krb_get_in_tkt (name, inst, realm, "krbtgt", realm, DEFAULT_TKT_LIFE, key_to_key, NULL, afskey) != 0)
	{
	  err = krb_get_in_tkt (name, inst, realm, "krbtgt", realm, DEFAULT_TKT_LIFE, key_to_key, NULL, mitkey);
	  if (err != 0)
	    {
	      syslog (LOG_ERR, "Unable to get tkt: %s",
		      krb_err_txt[err]);
	      memset (afskey, 0, sizeof (afskey));
	      memset (mitkey, 0, sizeof (mitkey));
	      return 1;
	    }
	}
    }

  /* All done with these */
  memset (afskey, 0, sizeof (afskey));
  memset (mitkey, 0, sizeof (mitkey));

#ifndef NO_AFS
  if ((err = krb_get_cred ("afs", "", realm, &c)) != 0)
    {
      if ((err = get_ad_tkt ("afs", "", realm, 255)) != 0)
	{
	  syslog (LOG_ERR, "Unable to get AFS service ticket: %s",
		  krb_err_txt[err]);
	  return 1;
	}
      else
	{
	  if ((err = krb_get_cred ("afs", "", realm, &c)) != 0)
	    {
	      syslog (LOG_ERR, "Unable to get AFS service ticket: %s",
		      krb_err_txt[err]);
	      return 1;
	    }
	}
    }

  strncpy (aserver.name, "afs", MAXKTCNAMELEN - 1);
  strncpy (aserver.instance, "", MAXKTCNAMELEN - 1);
  strncpy (aserver.cell, realm, MAXKTCREALMLEN - 1);

  {
    char *t = aserver.cell;
    char *s = aserver.cell;
    int c;
    while (c = *t++)
      {
	if (isupper (c))
	  c = tolower (c);
	*s++ = c;
      }
    *s++ = 0;
  }

  atoken.kvno = c.kvno;
  atoken.startTime = c.issue_date;
#ifdef AFS_LIFETIMES
  /* use AFS ticket lifetime lookup table */
  atoken.endTime = krb_life_to_time (c.issue_date, c.lifetime);
#else /* AFS_LIFETIMES */
  /* ticket lifetime is in five-minutes blocks. */
  atoken.endTime = c.issue_date + ((unsigned char) c.lifetime * 5 * 60);
#endif /* AFS_LIFETIMES */
  bcopy (c.session, &atoken.sessionKey, 8);
  atoken.ticketLen = c.ticket_st.length;
  bcopy (c.ticket_st.dat, atoken.ticket, atoken.ticketLen);

  if ((err = ktc_SetToken (&aserver, &atoken, &aclient)) != 0)
    {
      syslog (LOG_ERR, "Unable to set AFS token (%d)", err);
      return 1;
    }

#endif /* NO_AFS */
#else
  if (ka_UserAuthenticate (name, inst, /*realm */ 0, password, /*sepag */ 0, &reason))
    {
      syslog (LOG_ERR, "Unable to authenticate to AFS: %s", reason);
      return 1;
    }

#endif

  return 0;
}

int 
main (argc, argv)
     int argc;
     char *argv[];
{
  int len, nofork = 0, nap = 0, pid = -1, LogStatus, dosrvtab = 0, bkvno = 0;
  int pwfile = 0, dopidfile = 0;
  register int ch;
  unsigned tv;
  int arg;
  char password[64];   /* Should be big enough */
  char *name = NULL, *inst, *c, *filename, *pwfilename, lrealm[REALM_SZ];
  char *pidfilename = NULL;
#ifdef KERBEROS
  int get_key = 0;
#endif
  extern char *optarg;
  extern int optind;

  password[0] = '\0';
  krb_get_lrealm (lrealm, 1);

  /* need at least one argument to start with */
  if (argc == 1)
    {
      fprintf (stderr, "reauth: not enough arguments.\n");
      usage ();
    }

  openlog ("reauth", LOG_PID, LOGFACILITY);

  /* jeaton is less lazy than whoever wrote this before - we need a better
     argument parser here since we're adding more command line options */

  /* Okay, I'm still lazy.  Ripped from the old reauth code.  */
  for (arg = 1; arg < argc; arg++)
    {
      if (argv[arg][0] == '-')
	{
	  if (strcmp (argv[arg], "-nofork") == 0)
	    {
	      nofork = 1;
	    }
	  else if (strcmp (argv[arg], "-v") == 0)
	    {
	      verbose = 1;
	    }
	  else if (strcmp (argv[arg], "-verbose") == 0)
	    {
	      verbose = 1;
	    }
	  else if (strcmp (argv[arg], "-pw") == 0)
	    {
	      pwfile = 1;
	      pwfilename = argv[++arg];
	    }
	  else if (strcmp (argv[arg], "-srvtab") == 0)
	    {
	      dosrvtab = 1;
	      filename = argv[++arg];
	    }
	  else if (strcmp (argv[arg], "-pidfile") == 0)
	    {
	      dopidfile = 1;
	      pidfilename = argv[++arg];
	    }
	  else
	    {
	      fprintf (stderr, "reauth: unknown option %s\n", argv[arg]);
	      usage ();
	    }
	}
      else
	{			/* Not a dash arg */
#ifdef NOISY
	  printf ("not a dash arg: %s\n", argv[arg]);
#endif
	  if (nap == 0)
	    {
	      nap = atoi (argv[arg]);
	      if (nap <= 0)
		{
		  fprintf (stderr, "reauth: nap time must be greater than zero (%d)\n", nap);
		  usage ();
		}
#ifdef NOISY
	      printf ("nap time is %d\n", nap);
#endif
	    }
	  else if (name == NULL)
	    {
	      name = argv[arg];
	      if (c = strchr (name, '.'))
		{
		  *c = '\0';
		  inst = c + 1;
		}
	      else
		{
		  inst = "";
		}
#ifdef NOISY
	      printf ("name is %s\ninstance is %s\n", name, inst);
#endif
	    }
	  else if (password[0] == '\0')
	    {
	      len = strlen (argv[arg]);
	      if (len >= sizeof (password))
		{
		  fprintf (stderr, "reauth: password igiven on command line too long: %d > %d\n", len, sizeof (password));
		  usage ();
		}
	      strcpy (password, argv[arg]);
	      memset (argv[arg], '\0', len);
#ifdef NOISY
	      printf ("password read from command line\n");
#endif
	    }
	  else
	    {
	      fprintf (stderr, "reauth: too many non-dash arguments\n");
	      usage ();
	    }
	}
    }				/* End of argument parsing */

  /* Did we get the passwd? */
  if (password[0] == '\0')
    {
      if ((dosrvtab == 1) && (filename != NULL) && (filename != '\0'))
	{
	  /* Get it from srvtab */
	  memset (password, 0, sizeof (password));
	  read_service_key (name, inst, lrealm, bkvno, filename, (char *) password);
	}
      else if ((pwfile == 1) && (pwfilename != '\0'))
	{
	  if (verbose)
	    {
	      /* Get it from the file pwfilename */
	      printf ("reading from file %s", pwfilename);
	    }
	  pwfile = open (pwfilename, O_RDONLY);
	  read (pwfile, password, 64);
	  /* step on spurious newline */
	  if (c = strchr (password, '\n'))
	    {
	      *c = '\0';
	    }
	  close (pwfile);
	}
      else
	{
	  /* Get it from user now */
	  if (isatty (fileno (stdin)))
	    {
	      /* On a tty, so prompt for it */
	      if (des_read_pw_string (password, sizeof (password), "Password:", 0))
		{
		  password[0] = '\0';
		}
	    }
	  else
	    {
	      /* else just suck it in */
	      fgets (password, sizeof (password), stdin);
	      len = strlen (password);
	      /* step on spurious newline */
	      if (c = strchr (password, '\n'))
		{
		  *c = '\0';
		}
	    }
	}
    }

  /* Did we get the passwd? */
  if (password[0] == '\0')
    {
      fprintf (stderr, "reauth: no password given\n");
      usage ();
    }

  /* Now we have all the info we need */
  signal (SIGHUP, SIG_IGN);

  for (;;)
    {
      LogStatus = logme (name, inst, password);

      /* Fork & become daemon after first auth */
      if (pid == -1) {
	if(!nofork)
	{
	  pid = fork ();
	  if (pid < 0)
	    {
	      perror ("reauth");
	      fprintf (stderr, "reauth: fork failed\n");
	      syslog (LOG_ERR, "Unable to fork: %m");
	      exit (1);
	    }
	  if (pid > 0)
	    exit (0);
	  setpgrp (0, getpid ());
	}
 
        pid = getpid();

	if(dopidfile) {
	    FILE *pf;
	    pf = fopen(pidfilename, "w");
	    if(!pf) {
		fprintf(stderr, "could not open %s for writing\n",pidfilename);
		perror("error");
		exit(1);
	    }
	    fprintf(pf, "%d", pid);
	    fclose(pf);
	}
      }

      if (LogStatus == 0)
	{
	  tv = nap;
	}
      else
	{
	  /* only five minutes if we weren't able to authenticate */
	  /* unless nap time is less than five minutes, in wich case */
	  /* sleep for nap time */
	  if (nap < 300)
	    {
	      tv = nap;
	    }
	  else
	    {
	      tv = 5 * 60;
	    }
	}

#ifdef NOISY
      printf ("Sleeping %d\n", tv);
#endif
      if (verbose)
	{
	  syslog (LOGLEVEL, "sleeping %d", tv);
	}
      sleep (tv);
    }
}
